import argparse, warnings
from sbiRandM.sbiRandM import *
from sbiRandM.sbiRandM import superimp as sup


def parse_arguments():

    """
    Parsing of arguments. 
    -d : Path to the folder with the PDB's Pairwise Interactions that conforms the Complex.
    -fasta: Path to the Fasta File you want of the Complex that you want to model.
    -output: Path to the folder where generate the output file complex.
    -v: Activate the verbosity mode
    """

    parser = argparse.ArgumentParser(description='SbiRandM Complex Builder 1.0.0. This program generates a Protein-DNA or Protein-Protein complex, based on a Fasta Sequence and the pairwise interaction structures of the complex. This version uses a Superimposition-based approach to generate the structure.')

    parser.add_argument('-d', action="store", required =True, dest="folder", help="Path to the folder with the PDB's Pairwise Interactions that conforms the Complex.")
    parser.add_argument('-fasta', action="store", required=True, dest="fasta_seq", help="Path to the Fasta File you want of the Complex that you want to model.")
    parser.add_argument('-output', action="store", required=True, dest="output_folder", help="Path to the folder where generate the output file complex.")
    parser.add_argument('-v', '--verbose', dest = 'verbose', action = 'store_true', default = False, help = 'Activate the verbosity mode')

    args = vars(parser.parse_args())
    return args

args = parse_arguments()

sup.mainSuperimp(args)



import os, sys, warnings
from Bio.PDB.PDBParser import PDBParser
from Bio.PDB import PDBIO
import glob

warnings.filterwarnings("ignore")

def write_pairwise(pdb_file, pairwise_interactions, tmp):
    """
    This function generates pairwise interaction files, from a single Complex. Useful for testing.

    Input: PDB file, list of pairwise chain id that makes interactions, obtained with the function pairwise_generator
    Output: This function writes the PairWise interactions of chains in different PDB files in the same folder as input.
    """

    split_call = f'./modules/scripts/PDBtoSplitChain2.pl -i {pdb_file} -o {tmp}/temp_{os.path.basename(pdb_file)[:-4]}_' 
    os.system(split_call)
    for chainA, chainB in pairwise_interactions:
       ## !!! WARNING AÑADIR EL PDBTOADDCHAIN PARA PONER CADENAS A Y B EN LA PROTEINA 
        os.system(f'./modules/scripts/PDBtoAddChain.pl -i {tmp}/temp_{os.path.basename(pdb_file)[:-4]}_{chainA}.pdb -o {tmp}/temp_{os.path.basename(pdb_file)[:-4]}_{chainA}_mod -c "A"')
        os.system(f'./modules/scripts/PDBtoAddChain.pl -i {tmp}/temp_{os.path.basename(pdb_file)[:-4]}_{chainB}.pdb -o {tmp}/temp_{os.path.basename(pdb_file)[:-4]}_{chainB}_mod -c "B"')
        os.system(f'cat {tmp}/temp_{os.path.basename(pdb_file)[:-4]}_{chainA}_mod.pdb > {tmp}/complex_{chainA}_{chainB}.pdb')
        os.system(f'echo "TER" >> {tmp}/complex_{chainA}_{chainB}.pdb')
        os.system(f'cat {tmp}/temp_{os.path.basename(pdb_file)[:-4]}_{chainB}_mod.pdb >> {tmp}/complex_{chainA}_{chainB}.pdb')
    filelist = glob.glob(f"{tmp}/temp_*.*")
    for f in filelist:
        os.remove(f)


def check_chain_interaction(chainA, chainB, threshold):
    """
    This chain takes two BioPython PDB Chain objects, and return True is any atom CB of the first chain
    is closer than threshold to one atom CB of the second chain. Otherwise, returns False.
    """
    if chainA == chainB:
        #print ("Chains are equal, skipping")
        return False
    else:
        for residue_A in chainA:
            for residue_B in chainB:
                for atom_A in residue_A:
                    for atom_B in residue_B:
                        #if atom_A.id == "CB" and atom_B.id == "CB":
                            if atom_A - atom_B < threshold: 
                                print ("Chains interact")
                                return True
                            else: continue
        return False
       

def pairwise_generator(pdb_file, tmp):
    """
    This function takes a PDB file, and decompose it into the chains objects of Biopython.
    Then it calls the function check_chain_interactions, and returns a list of tupples containing the ID of the chains that interact.
    """

    #io = PDBIO()
    parser = PDBParser(PERMISSIVE=1)
    interacting_chains = list()
    structure = parser.get_structure('Complex', pdb_file)
    for model in structure:
        if len(model) == 1:
            return ("PDB only has one chain.")
        else:
            #print ("PDB has the next number of chains:", len(model))
            for chain in model:
                for second_chain in model:
                    if check_chain_interaction(chain, second_chain, 20): interacting_chains.append((chain.id, second_chain.id))

            unique_pairwise_interactions = ([v for k, v in enumerate(interacting_chains) if v[::-1] not in interacting_chains[:k]])
            write_pairwise(pdb_file, unique_pairwise_interactions, tmp)
                        
    return None

'''
if __name__ == "__main__":
    pdb_path = os.path.join(os.path.dirname(os.path.abspath(sys.argv[1])))
    pdb_file = os.path.basename(sys.argv[1])
    pdb_full_path = os.path.join(pdb_path, pdb_file)
    try:
        TMP_folder = os.path.join(pdb_path, "TMP")
        os.mkdir(TMP_folder)
    except:
        print ("Directory is already created, skipping.")

    print (pdb_file)
    print(pairwise_generator(pdb_full_path, TMP_folder))

'''


import sys, os
from Bio import SeqIO
from Bio import pairwise2
import itertools, difflib
from sbiRandM.sbiRandM.data import Alphabet

def check_fasta_stoichometry(fasta_path):

    """
    This function takes the path of the Fasta file of the Complex, and checks the homology between chains
    to detect the steichiometry of the protein.
    
    @input - Fasta file of the Complex

    @output - Dictionary with the steichiometry of the complex.
              Name of the chain : 
                        Steichiometry - Absolute number that the chain appear in the Fasta.
                        Sequence - Sequence of the chain
    """

    steichiometry_dict = dict()
    chains = list(SeqIO.parse(fasta_path, "fasta"))
    index = 0

    for chain in chains:

        chain.id = Alphabet[index]
        steichiometry_dict[chain.id] = {"steichiometry" : 1, "sequence" : str(chain.seq)}
        index += 1

    for a,b in itertools.combinations(chains, 2):

        if difflib.SequenceMatcher(None, str(a.seq), str(b.seq)).ratio() > 0.70:
            #print ("Chains", a.id, b.id, "match!")
            try:
                steichiometry_dict[a.id]["steichiometry"] += 1
                del steichiometry_dict[b.id]
            except:
                continue
    
    return steichiometry_dict

if __name__ == "__main__":
    print( check_fasta_stoichometry(sys.argv[1]) )




import sys, os, warnings
from Bio.PDB import PDBIO
from Bio.PDB import PDBParser
from Bio.PDB import Superimposer
from Bio import SeqIO
from sbiRandM.sbiRandM import *
from sbiRandM.sbiRandM.exceptions import *

from sbiRandM.sbiRandM.data import Alphabet
import itertools, random
from shutil import copyfile


def Compute_equal_chain(structure_1, structure_2):
    
    """
    @ Input - Two Biopython PDB structures of an interaction.
    @ Output - A list of Atoms that correspond to the same chain.
    """

    for chain in structure_1[0]:
        for chain_2 in structure_2[0]:
            if Get_fasta(chain) == Get_fasta(chain_2):
                return (list(chain.get_atoms()), list(chain_2.get_atoms()))

    return (None, None)

def check_chain_addition(complex_pdb, chain, pairwise_dict, steichiometry_dict):
    """
    TO-DO
    Esto tiene que mirar si el complex PDB y la chain.id que queremos añadir, tienen alguna interaccion posible
    si la tienen, se devuelve el path del file
    si no la tiene, se devuelve false, por lo que se hace otro pick at random
    """
    parser = PDBParser(PERMISSIVE=1)

    complex_pdb = parser.get_structure('Complex', complex_pdb)
    
    try:
        chains_to_test_in_complex = list(pairwise_dict[chain].keys())
    except KeyError:
        raise FastaRaroException
    common_chains = list()
    
    
    for chain_pdb in complex_pdb.get_chains():
        fasta = Get_fasta(chain_pdb)
        for chain_2 in chains_to_test_in_complex:
            if check_homology(fasta, steichiometry_dict[chain_2]["sequence"]):
                common_chains.append(pairwise_dict[chain][chain_2])
    
    return common_chains

    return None

def build_complex(file_1, file_2):
    """
    This function takes the complex output file (or in the first iteration one of the pairwise interactions)
    and another pairwise interaction PDB complex. Then it tries to add the chain to the complex until there is not clash

    @ Input - Two file path for a PDB interactions.
    @ Output - File path of the complex PDB file / Error: Chain cannot be added.
    """

    parser = PDBParser(PERMISSIVE=1)

    structure_1 = parser.get_structure('Complex', file_1)
    structure_2 = parser.get_structure('Complex', file_2)

    sup = Superimposer()
    io = PDBIO()

    atoms_fixed, atoms_moving = Compute_equal_chain(structure_1, structure_2)

    try:
        sup.set_atoms(atoms_fixed, atoms_moving)
    except:
        return False

    sup.apply(list(structure_2.get_atoms()))

    for chain in structure_2[0].get_chains():
        if chain.id != list(atoms_moving)[0].get_full_id()[2]:
            moved_chain = chain

    if check_clash(structure_1, moved_chain):
        with open(file_1, "wt") as out_file:
            
            for model in list(structure_1.get_chains()) + [moved_chain] :
                io.set_structure(model)
                io.save(out_file)

        rename_complex_chains(file_1)

        return True
    return False


def rename_complex_chains(file):
    """
    This function takes a PDB file, delete all the END lines, and rename the chains.
    according to Alphabet found in data module.

    @ Input - PDB file
    @ Output - PDB file with chains named.
    """

    index = 0
    temp_file = os.path.join(os.path.dirname(file),"temporal.pdb")
    with open(file, "r") as input_file:
        with open(temp_file,"w") as out_file:

            for line in input_file:
                if "TER" in line:
                    index += 1
                    continue
                else:
                    if "ATOM" in line:
                        line = list(line)
                        line[21] = Alphabet[index]
                        out_file.write("".join(line))
    os.remove(file)
    os.rename(temp_file, file)

def execute_complex(steichiometry_dict, pairwise_dict, args):

    chains_list = [[i] * steichiometry_dict[i]["steichiometry"] for i  in steichiometry_dict]
    chains_list = list(itertools.chain.from_iterable(chains_list))

    random_start = random.sample(chains_list, 2)
    random_start_1 = random.choice(list(pairwise_dict.keys()))
    random_start_2 = random.choice(list(pairwise_dict[random_start_1].keys()))

    if args['verbose']:
        print ("Starting from random chains:", random_start[0] ,"and", random_start[1])

    start_complex = pairwise_dict[random_start_1][random_start_2]
    pdb_complex = os.path.join(args['output_folder'], "complex.pdb")
    copyfile(start_complex, pdb_complex)

    chains_list.remove(random_start_1)
    chains_list.remove(random_start_2)

    while len(chains_list) > 0:
        try_chain = random.choice(chains_list)

        file_to_try = []

        while len(file_to_try) == 0:

            try_chain = random.choice(chains_list)
            file_to_try = check_chain_addition(pdb_complex , try_chain , pairwise_dict , steichiometry_dict)

        print("Trying to add" , try_chain)

        for file_pdb in file_to_try:
            chain_addition = build_complex(pdb_complex , file_pdb)

            if chain_addition:
                chains_list.remove(try_chain)
                print("Chain added successfully. Number of remaining chains:" , len(chains_list))
                if len(chains_list) == 0:
                    print("All chains have been added.")
                elif args['verbose']:
                    if len(chains_list) != 0:
                        print("The remaining chains to add are" , chains_list)
                break
    
    return os.path.join(args['output_folder'], "superimposition_complex.pdb")




from .sbiRandM.exceptions import *
from .sbiRandM.superimp import *
from .sbiRandM.check_stoichometry import *
from .sbiRandM.interaction_module import *
from .sbiRandM.data import *
from .sbiRandM.GUI.mainWindow import *
from .sbiRandM.models import *






#!/usr/bin/env python3
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="sbiRandM",
    version="1.0.4",
    author="Ruben-And-Miguel",
    author_email="ruben.molina-fernandez@upf.edu",
    description="Protein-DNA Complex builder",
    long_description="This is the Structural Bioinformatics project of Ruben Molina and Miguel Dieguez.",
    long_description_content_type="text/markdown",
    url="https://github.com/RMolina93/Structural_Project",
    packages=['sbiRandM','sbiRandM.sbiRandM'],
    install_requires=['biopython'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



class FilesDontMatchException(Exception):

   def __init__(self): 
      pass
   def __str__(self):
      print ("The program has stopped because the files do not match the requirements.")
      return "The files do not match the requirements"

class FastaRaroException(Exception):

   def __init__(self): 
      pass
   def __str__(self):
      print ("The program stopped because the fasta file does not match with the files in the provided PDB folder")
      return "The fasta file does not match with the pdb directory provided"
   
class BadFastaException(Exception):

   def __init__(self): 
      pass
   def __str__(self):
      print ("The program stopped because your Fasta file is either bad configured or have loosen information.")
      return "Your fasta file is either bad configured or it's likely to have loosen information"